from setuptools import setup, find_packages

setup(name="spylang",
      author='Antonio Cuni',
      author_email='anto.cuni@gmail.com',
      url='https://github.com/antocuni/spylang',
      packages=find_packages(),
      zip_safe=False,
      )
